/*
 * Copyright (c) University of Toronto 1990.
 * All rights reserved.
 * Written by Mark Moraes.
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or of the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The authors are not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */
/*
 * find() walks the file system starting from the directory path,
 * invoking proc() on every directory entry it finds.  proc is invoked
 * with the pathname of the entry and the stat structure resulting from
 * an lstat on that file.  (The only exception is if the path first
 * supplied to find() ends with a "/", in which case a stat() is
 * performed on it.  That is the only case where find() will follow a
 * symlink if it encounters one).  find() returns -1 if an error
 * occurred, 0 if it succeeded, or the negative number returned by the
 * last (*proc)() invocation.  If (*proc)() returns a positive number,
 * then find() will not descend into the entry on which this happens,
 * but will continue the walk.
 */
/*
 * !! find() can run out of file descriptors.  If a chdir() fails, this
 * exits with a fatal error.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#ifdef SYS_DIR
# include <sys/dir.h>
# define dirent direct
#else
# include <dirent.h>
#endif
#include "libc.h"

#ifndef MAXPATHLEN
# ifdef PATH_MAX
#  define MAXPATHLEN PATH_MAX
# else
#  define MAXPATHLEN 1024
# endif
#endif
#define MAXSTR	128

/* forward declarations */
static int dofind();

void
echdir(dir, cwd)
char *dir;
char *cwd;
{
	if(chdir(dir) < 0) {
		char ebuf[MAXSTR];
		(void) sprintf(ebuf, "chdir(%s) failed, cwd = %s", dir, cwd);
		error("%s", ebuf);
		/*NOTREACHED*/
	}
}


static char curdir[MAXPATHLEN];
static char *cdirp = curdir;

int
find(path, proc)
char *path;
int (*proc)();
{
	char cwd[MAXPATHLEN];
	int i;
	if (getcwd(cwd, MAXPATHLEN) == 0) {
		warning("getcwd failed: %s", cwd);
		return -1;
	}
	i = dofind(path, proc);
	echdir(cwd, "");
	return i;
}
		
static int
dofind(path, proc)
char *path;
int (*proc)();
{
	DIR *dirp;
	struct dirent *dp;
	struct stat stbuf;
	char ebuf[MAXSTR];
	char *fname;
	char *ocdirp;
	char c;
	int i;
	int (* statproc)();

	i = strlen(path) - 1;
	if (path[i] == '/') {
		statproc = stat;
		path[i] = '\0';
	} else
		statproc = lstat;
	if ((* statproc)(path, &stbuf) == -1) {
		sprintf(ebuf, "%sstat(%s) failed, cwd = %s",
			statproc == lstat? "l" : "", path, curdir);
		warning("%s", ebuf);
		return -1;
	}

	ocdirp = cdirp;
	(void) strcpy(cdirp, path);
	cdirp += strlen(path);
	*cdirp = '\0';

	if ((i = (* proc)(curdir, &stbuf)) != 0) {
		if (i > 0)	/* don't descend, but keep going */
			i = 0;
		goto ret2;
	}
	if ((stbuf.st_mode & S_IFMT) != S_IFDIR) {
		i = 0;
		goto ret2;
	}

	/* restore curdir for the error message, if any */
	c = *ocdirp;
	*ocdirp = '\0';
	echdir(path, curdir);
	dirp = opendir(".");
	if (dirp == NULL) {
		warning("opendir(\".\") failed, path = %s", curdir);
		i = -1;
		goto ret;
	}
	*ocdirp = c;
	*cdirp++ = '/';
	*cdirp = '\0';

	i = 0;
	while((dp = readdir(dirp)) != NULL) {
		fname = dp->d_name;
		if (*fname == '.' && (*(fname + 1) == '\0' ||
				      (*(fname + 1) == '.' &&
				       *(fname + 2) == '\0')))
			continue;
		if ((i = dofind(fname, proc)) > 0)
			break;
	}
	(void) closedir(dirp);
ret:
	if (ocdirp != curdir)
		echdir("..", curdir);
ret2:
	cdirp = ocdirp;
	*cdirp = '\0';
	return i;
}
