/*
 * Reliable locking via files and link().  Adapted from the code in C
 * news.  It will automatically override stale locks on the same machine,
 * but will not remove locks from other machines since it has no way of
 * determining staleness.
 */
/* Mark Moraes, University of Toronto */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include "libc.h"

/* define this to compile standalone for testing */
/* #define TESTLOCK */

#define LOCKTEMP "LOCKTMXXXXXX"
#define MAXHOSTSTR 128
#define MAXPIDSTR 32

static FILE *debugfp = 0;
static char *ourhname = NULL;
static unsigned int ourhlen;

/* imports */
extern char *progname;

/* forward declarations */
static int stalelock();

void
lockdebug(fp)
FILE *fp;
{
	debugfp = fp;
}

/* 
 * Try to lock a file.  The lock directory must be writable -- we create
 * a temporary name there, store our hostname and pid in it, repeatedly
 * try to link the temporary name to lockname.
 */
int
lockcreate(lockname, lockretries, interval)
char *lockname;
int lockretries;
unsigned int interval;
{
	char *tempnm;
	int tempfd;
	unsigned int len;
	char *cp;
	char buf[MAXPIDSTR];		/* big enough for a pid */
	char hname[MAXHOSTSTR];

	cp = strrchr(lockname, '/');
	if (cp == NULL) {
		tempnm = strsave(LOCKTEMP);
	} else {
		len = cp - lockname + 1;	/* includes the '/' */
		tempnm = emalloc(len + sizeof(LOCKTEMP));
		(void) memcpy(tempnm, lockname, len);
		(void) memcpy(tempnm + len, LOCKTEMP, sizeof(LOCKTEMP));
	}
	tempfd = opentmp(tempnm);
	if (tempfd < 0)
		error("can't opentmp(%s)", tempnm);
	if (ourhname == NULL) {
		cp = hname + sizeof(hname) - 2;
		*cp = '\0';
		if (gethostname(hname, sizeof(hname) - 1) < 0)
			error("gethostname failed.", "");
		if (*cp != '\0')
			error("hostname too large.", "");
		(void) strcat(hname, " ");
		ourhlen = strlen(hname);
		ourhname = strcpy((char *) emalloc(ourhlen + 1), hname);
	}
	(void) sprintf(buf, "%d\n", getpid());
	len = strlen(buf);
	if (write(tempfd, ourhname, ourhlen) != ourhlen ||
	    write(tempfd, buf, len) != len ||
	    close(tempfd) < 0)
		error("can't write/close(%s)", tempnm);
	while (lockretries >= 0 && link(tempnm, lockname) < 0) {
		if (errno != EEXIST)
			warning("can't link lock tempfile to `%s'", lockname);
		if (!stalelock(lockname)) {
			if (debugfp)
				(void) fprintf(debugfp, "%s: sleeping on %s\n",
					       progname, lockname);
			sleep(interval);
		}
		lockretries--;
	}
	(void) unlink(tempnm);
	free(tempnm);
	if (debugfp)
		(void) fprintf(debugfp, "%s: %s %s\n", progname,
			       (lockretries >= 0) ? "locked" : "can't lock",
			       lockname);
	return (lockretries >= 0) ? 0 : -1;
}

void
lockfree(lockname)
char *lockname;
{
	(void) unlink(lockname);
	if (debugfp)
		(void) fprintf(debugfp, "%s: unlocked %s\n", progname,
			       lockname);
}

/*
 * if a lock is stale, removes it.  Returns 1 if it clears the lock, 0
 * otherwise.  A lock is stale if the hostname stored in it is host name
 * and kill(lockpid, 0) returns -1.  If it isn't our hostname, then we
 * can't decide if it is stale, so we leave it alone.  Since the lock
 * could vanish under us at any time, the failure of an fopen or read
 * isn't considered fatal.
 */
static int
stalelock(lockname)
char *lockname;
{
	FILE *tempfp;
	char hbuf[MAXHOSTSTR + MAXPIDSTR + 2];
	int lockpid;
	
	if ((tempfp = fopen(lockname, "r")) == NULL)
		return 0;
	if (fgets(hbuf, sizeof(hbuf), tempfp) == NULL ||
	    strncmp(hbuf, ourhname, ourhlen) != 0) {
		(void) fclose(tempfp);
		return 0;
	}
	lockpid = atoi(hbuf + ourhlen);
	if (kill(lockpid, 0) < 0) {
		if (debugfp)
			(void) fprintf(debugfp,
				"%s: removing stale lock once held by %s:%d\n",
			       progname, ourhname, lockpid);
		(void) fclose(tempfp);
		lockfree(lockname);
		return 1;
	}
	return 0;
}

#ifdef TESTLOCK
char *
emalloc(n)
unsigned n;
{
	char *cp = malloc(n);

	if (cp == NULL)
		error("out of memory", "");
	return cp;
}

char *
strsave(s)
char *s;
{
	return strcpy(emalloc(strlen(s) + 1), s);
}

char *progname = "tlock";

void
error(fmt, s)
char *fmt, *s;
{
	if (errno)
		perror(progname);
	(void) fprintf(stderr, fmt, s);
	fputc('n', stderr);
}

int
main(argc, argv)
int argc;
char **argv;
{
	char **cpp;
	lockdebug(stderr);
	for(cpp = argv + 1; *cpp; cpp++)
		lockcreate(*cpp, 5, (unsigned) 1);
	sleep((unsigned) 60);
	for(cpp = argv + 1; *cpp; cpp++)
		lockfree(*cpp);
	return 0;
}
#endif
