/*
 * Copyright (c) University of Toronto 1990.
 * All rights reserved.
 * Written by Ken Lalonde.
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or of the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The authors are not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */
/*
 * Cacheing replacements for getpwnam, getpwuid and the /etc/group equivalents.
 * The passwd and group lists are read into a binary search tree.
 * This is not necessary if the libc versions use a fast
 * lookup scheme (e.g. 4.3BSD), but it is a great speedup if they don't.
 *
 * The group membership list, gr->gr_mem, is ignored.  If you need it,
 * compile -DNEED_GRMEM.
 *
 * Note: this version does lazy evaluation: it simply caches the results
 * of the libc versions.  This is a win if the calling program
 * requests a small set of keys many times (as track does).
 * If the key space is large, it's better to read everything in at once
 * by calling pgcache_init().
 */
/* LINTLIBRARY */
#include <stdio.h>
#include <sys/types.h>
#include <varargs.h>
#include <search.h>
#include "libc.h"
#include "pgcache.h"

static void addpw(), addgr(), add();
static struct passwd *pw_flatten();
static struct group *gr_flatten();
static char *lookup(), *format();
static char null;	/* dummy data address for missing keys */


struct passwd *
getpwuid_cache(uid)
{
	struct passwd *pw = (struct passwd *)lookup(format("p#%d", uid));
	if (pw == NULL) {
		if ((pw = getpwuid(uid)) == NULL)
			add(format("p#%d", uid), &null);
		else
			addpw(pw);
	}
	return pw == (struct passwd *)&null ? NULL : pw;
}

struct passwd *
getpwnam_cache(name)
	char *name;
{
	struct passwd *pw = (struct passwd *)lookup(format("pn%s", name));
	if (pw == NULL) {
		if ((pw = getpwnam(name)) == NULL)
			add(format("pn%s", name), &null);
		else
			addpw(pw);
	}
	return pw == (struct passwd *)&null ? NULL : pw;
}

struct group *
getgrgid_cache(gid)
{
	struct group *gr = (struct group *)lookup(format("g#%d", gid));
	if (gr == NULL) {
		if ((gr = getgrgid(gid)) == NULL)
			add(format("g#%d",gid), &null);
		else
			addgr(gr);
	}
	return gr == (struct group *)&null ? NULL : gr;
}

struct group *
getgrnam_cache(name)
	char *name;
{
	struct group *gr = (struct group *)lookup(format("gn%s", name));
	if (gr == NULL) {
		if ((gr = getgrnam(name)) == NULL)
			add(format("gn%s",name), &null);
		else
			addgr(gr);
	}
	return gr == (struct group *)&null ? NULL : gr;
}

void
pgcache_init()
{
	struct passwd *pw;
	struct group *gr;

	setpwent();
	while (pw = getpwent())
		addpw(pw);
	endpwent();

	setgrent();
	while (gr = getgrent())
		addgr(gr);
	endgrent();
}

/*
 * Internals.
 */

struct node {
	char *key, *data;
};
static struct node *root;	/* root of binary search tree */
static char buf[BUFSIZ];	/* various uses */

static int
tcmp(n1, n2)
	struct node *n1, *n2;
{
	return strcmp(n1->key, n2->key);
}

static char *
format(va_alist)	/* does sprintf(buf, ...) */
	va_dcl
{
	va_list ap;
	char *fmt;

	/* Build the search key */
	va_start(ap);
	fmt = va_arg(ap, char *);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	return buf;
}

static char *
lookup(key)
	char *key;
{
	struct node k, **e;

	k.key = key;
	e = (struct node **)tfind((void *) &k, (void **) &root, tcmp);
	return e ? (*e)->data : NULL;
}

static void
addpw(pw)
	struct passwd *pw;
{
	pw = pw_flatten(pw);	/* make contiguous */
	if (pw == NULL)
		return;
	add(format("pn%s", pw->pw_name), (char *)pw);
	add(format("p#%d", pw->pw_uid),  (char *)pw);
}

static void
addgr(gr)
	struct group *gr;
{
	gr = gr_flatten(gr);
	if (gr == NULL)
		return;
	add(format("gn%s", gr->gr_name), (char *)gr);
	add(format("g#%d", gr->gr_gid),  (char *)gr);
}

/*
 * Add an entry to the tree.
 * The entry has 2 keys, one for searching by number, one by name.
 * The keys look like "type#N" and "typenNAME" respectively.
 */
static void
add(key, data)
	char *key;
	char *data;
{
	struct node *k;
	char *p;

	p = malloc((sizeof *k) + strlen(key)+2);
	if (p == NULL)
		return;
	k = (struct node *)p;
	p = (char *)(k+1);
	strcpy(k->key = p, key);
	k->data = data;
	(void) tsearch((void *)k, (void **) &root, tcmp);
}

/*
 * Copy a passwd structure to new contiguous space.
 */
static struct passwd *
pw_flatten(old)
	struct passwd *old;
{
	struct passwd *new;
	char *p;

	new = (struct passwd *)malloc(sizeof *old +
		strlen(old->pw_name)+1 +
		strlen(old->pw_passwd)+1 +
		strlen(old->pw_gecos)+1 +
		strlen(old->pw_dir)+1 +
		strlen(old->pw_shell)+1);
	if (new == NULL)
		return NULL;
	new->pw_uid = old->pw_uid;
	new->pw_gid = old->pw_gid;
	p = (char *)(new+1);
#define z(f)  strcpy(new->f = p, old->f); p += strlen(p)+1;
	z(pw_name);
	z(pw_passwd);
	z(pw_gecos);
	z(pw_dir);
	z(pw_shell);
	return new;
}

static struct group *
gr_flatten(old)
	struct group *old;
{
	struct group *new;
	char *p, **np;
	int strsize = 0;	/* string space for gr_mem elements */
	int ptrsize = 0;	/* array of ptrs to gr_mem + final null ptr */

#ifdef NEED_GRMEM
	char **op;

	for (op = old->gr_mem; *op; op++)
		ptrsize += sizeof(char *), strsize += strlen(*op)+1;
	ptrsize += sizeof(char *);	/* final null */
#endif
	new = (struct group *) malloc(sizeof *old +
		ptrsize +
		strsize +
		strlen(old->gr_name)+1 +
		strlen(old->gr_passwd)+1);
	if (new == NULL)
		return NULL;
	new->gr_gid = old->gr_gid;
	new->gr_mem = np = (char **)(new+1);
#ifdef NEED_GRMEM
	p = (char *)np + ptrsize;
	for (op = old->gr_mem; *op; op++) {
		*np++ = p;
		strcpy(p, *op);
		p += strlen(p)+1;
	}
	*np++ = NULL;
#endif
	p = (char *)np;
	z(gr_name);
	z(gr_passwd);
	return new;
}
